#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""Output formatters using prettytable."""

import argparse
import collections.abc
import os
import sys
import typing as ty

import prettytable

from cliff import columns
from cliff.formatters import base
from cliff import utils

_T = ty.TypeVar('_T')


def _format_row(
    row: collections.abc.Iterable[
        columns.FormattableColumn[ty.Any] | str | _T
    ],
) -> list[_T | str]:
    new_row = []
    for r in row:
        if isinstance(r, columns.FormattableColumn):
            r = r.human_readable()
        if isinstance(r, str):
            r = r.replace('\r\n', '\n').replace('\r', ' ')
        new_row.append(r)
    return new_row


def _do_fit(fit_width: bool) -> bool:
    if os.name == 'nt':
        # NOTE(pas-ha) the isatty is not reliable enough on Windows,
        # so do not try to auto-fit
        return fit_width
    return sys.stdout.isatty() or fit_width


class TableFormatter(base.ListFormatter, base.SingleFormatter):
    ALIGNMENTS = {
        int: 'r',
        str: 'l',
        float: 'r',
    }

    def add_argument_group(self, parser: argparse.ArgumentParser) -> None:
        group = parser.add_argument_group('table formatter')
        group.add_argument(
            '--max-width',
            metavar='<integer>',
            default=int(os.environ.get('CLIFF_MAX_TERM_WIDTH', 0)),
            type=int,
            help=(
                'Maximum display width, <1 to disable. You can also '
                'use the CLIFF_MAX_TERM_WIDTH environment variable, '
                'but the parameter takes precedence.'
            ),
        )
        group.add_argument(
            '--fit-width',
            action='store_true',
            default=bool(int(os.environ.get('CLIFF_FIT_WIDTH', 0))),
            help=(
                'Fit the table to the display width. '
                'Implied if --max-width greater than 0. '
                'Set the environment variable CLIFF_FIT_WIDTH=1 '
                'to always enable'
            ),
        )
        group.add_argument(
            '--print-empty',
            action='store_true',
            help='Print empty table if there is no data to show.',
        )

    def add_rows(
        self,
        table: prettytable.PrettyTable,
        column_names: collections.abc.Sequence[str],
        data: collections.abc.Iterable[collections.abc.Sequence[ty.Any]],
    ) -> None:
        # Figure out the types of the columns in the
        # first row and set the alignment of the
        # output accordingly.
        data_iter = iter(data)
        try:
            first_row = next(data_iter)
        except StopIteration:
            pass
        else:
            for value, name in zip(first_row, column_names):
                alignment = self.ALIGNMENTS.get(type(value), 'l')
                table.align[name] = alignment
            # Now iterate over the data and add the rows.
            table.add_row(_format_row(first_row))
            for row in data_iter:
                table.add_row(_format_row(row))

    def emit_list(
        self,
        column_names: collections.abc.Sequence[str],
        data: collections.abc.Iterable[collections.abc.Sequence[ty.Any]],
        stdout: ty.TextIO,
        parsed_args: argparse.Namespace,
    ) -> None:
        x = prettytable.PrettyTable(
            column_names,
            print_empty=parsed_args.print_empty,
        )
        x.padding_width = 1

        # Add rows if data is provided
        if data:
            self.add_rows(x, column_names, data)

        # Choose a reasonable min_width to better handle many columns on a
        # narrow console. The table will overflow the console width in
        # preference to wrapping columns smaller than 8 characters.
        min_width = 8
        self._assign_max_widths(
            x, int(parsed_args.max_width), min_width, parsed_args.fit_width
        )

        formatted = x.get_string()
        stdout.write(formatted)
        stdout.write('\n')
        return

    def emit_one(
        self,
        column_names: collections.abc.Sequence[str],
        data: collections.abc.Sequence[ty.Any],
        stdout: ty.TextIO,
        parsed_args: argparse.Namespace,
    ) -> None:
        x = prettytable.PrettyTable(
            field_names=('Field', 'Value'), print_empty=False
        )
        x.padding_width = 1
        # Align all columns left because the values are
        # not all the same type.
        x.align['Field'] = 'l'
        x.align['Value'] = 'l'
        for name, value in zip(column_names, data):
            x.add_row(_format_row((name, value)))

        # Choose a reasonable min_width to better handle a narrow
        # console. The table will overflow the console width in preference
        # to wrapping columns smaller than 16 characters in an attempt to keep
        # the Field column readable.
        min_width = 16
        self._assign_max_widths(
            x, int(parsed_args.max_width), min_width, parsed_args.fit_width
        )

        formatted = x.get_string()
        stdout.write(formatted)
        stdout.write('\n')
        return

    @staticmethod
    def _field_widths(
        field_names: list[str], first_line: str
    ) -> dict[str, int]:
        # use the first line +----+-------+ to infer column widths
        # accounting for padding and dividers
        widths = [max(0, len(i) - 2) for i in first_line.split('+')[1:-1]]
        return dict(zip(field_names, widths))

    @staticmethod
    def _width_info(term_width: int, field_count: int) -> tuple[int, int]:
        # remove padding and dividers for width available to actual content
        usable_total_width = max(0, term_width - 1 - 3 * field_count)

        # calculate width per column if all columns were equal
        if field_count == 0:
            optimal_width = 0
        else:
            optimal_width = max(0, usable_total_width // field_count)

        return usable_total_width, optimal_width

    @staticmethod
    def _build_shrink_fields(
        usable_total_width: int,
        optimal_width: int,
        field_widths: dict[str, int],
        field_names: list[str],
    ) -> tuple[list[str], int]:
        shrink_fields = []
        shrink_remaining = usable_total_width
        for field in field_names:
            w = field_widths[field]
            if w <= optimal_width:
                # leave alone columns which are smaller than the optimal width
                shrink_remaining -= w
            else:
                shrink_fields.append(field)

        return shrink_fields, shrink_remaining

    @staticmethod
    def _assign_max_widths(
        table: prettytable.PrettyTable,
        max_width: int,
        min_width: int = 0,
        fit_width: bool = False,
    ) -> None:
        """Set maximum widths for columns of table `x`,
        with the last column recieving either leftover columns
        or `min_width`, depending on what offers more space.
        """
        if max_width > 0:
            term_width = max_width
        elif not _do_fit(fit_width):
            # Fitting is disabled
            return None
        else:
            _term_width = utils.terminal_width()
            if not _term_width:
                # not a tty, so do not set any max widths
                return None
            term_width = _term_width
        field_count = len(table.field_names)

        try:
            first_line = table.get_string().splitlines()[0]
            if len(first_line) <= term_width:
                return None
        except IndexError:
            return None

        usable_total_width, optimal_width = TableFormatter._width_info(
            term_width, field_count
        )

        field_widths = TableFormatter._field_widths(
            table.field_names, first_line
        )

        shrink_fields, shrink_remaining = TableFormatter._build_shrink_fields(
            usable_total_width, optimal_width, field_widths, table.field_names
        )

        shrink_to = shrink_remaining // len(shrink_fields)
        # make all shrinkable fields size shrink_to apart from the last one
        for field in shrink_fields[:-1]:
            table.max_width[field] = max(min_width, shrink_to)
            shrink_remaining -= shrink_to

        # give the last shrinkable column any remaining shrink or min_width
        field = shrink_fields[-1]
        table.max_width[field] = max(min_width, shrink_remaining)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""Output formatters using PyYAML."""

import argparse
import collections.abc
import typing as ty

from cliff import columns
from cliff.formatters import base


def _yaml_friendly(value: ty.Any) -> ty.Any:
    if isinstance(value, columns.FormattableColumn):
        return value.machine_readable()
    elif hasattr(value, "toDict"):
        return value.toDict()
    elif hasattr(value, "to_dict"):
        return value.to_dict()
    else:
        return value


class YAMLFormatter(base.ListFormatter, base.SingleFormatter):
    def add_argument_group(self, parser: argparse.ArgumentParser) -> None:
        pass

    def emit_list(
        self,
        column_names: collections.abc.Sequence[str],
        data: collections.abc.Iterable[collections.abc.Sequence[ty.Any]],
        stdout: ty.TextIO,
        parsed_args: argparse.Namespace,
    ) -> None:
        # the yaml import is slow, so defer loading until we know we want it
        import yaml

        items = []
        for item in data:
            items.append(
                {n: _yaml_friendly(i) for n, i in zip(column_names, item)}
            )
        yaml.safe_dump(items, stream=stdout, default_flow_style=False)

    def emit_one(
        self,
        column_names: collections.abc.Sequence[str],
        data: collections.abc.Sequence[ty.Any],
        stdout: ty.TextIO,
        parsed_args: argparse.Namespace,
    ) -> None:
        # the yaml import is slow, so defer loading until we know we want it
        import yaml

        for key, value in zip(column_names, data):
            dict_data = {key: _yaml_friendly(value)}
            yaml.safe_dump(dict_data, stream=stdout, default_flow_style=False)



import logging
import os

from cliff.lister import Lister


class Files(Lister):
    """Show a list of files in the current directory.

    The file name and size are printed by default.
    """

    log = logging.getLogger(__name__)

    def take_action(self, parsed_args):
        return (
            ('Name', 'Size'),
            ((n, os.stat(n).st_size) for n in os.listdir('.')),
        )



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""Application base class for providing a list of data as output."""

import abc
import argparse
import collections.abc
import logging
import typing as ty

from cliff import _argparse
from cliff import display
from cliff.formatters import base as base_formatters


class Lister(
    display.DisplayCommandBase[base_formatters.ListFormatter],
    metaclass=abc.ABCMeta,
):
    """Command base class for providing a list of data as output."""

    log = logging.getLogger(__name__)

    @property
    def formatter_namespace(self) -> str:
        return 'cliff.formatter.list'

    @property
    def formatter_default(self) -> str:
        return 'table'

    @property
    def need_sort_by_cliff(self) -> bool:
        """Whether sort procedure is performed by cliff itself.

        Should be overridden (return False) when there is a need to implement
        custom sorting procedure or data is already sorted.
        """
        return True

    @abc.abstractmethod
    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[
        collections.abc.Sequence[str], collections.abc.Iterable[ty.Any]
    ]:
        """Run command.

        Return a tuple containing the column names and an iterable containing
        the data to be listed.
        """

    def get_parser(self, prog_name: str) -> _argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        group = self._formatter_group
        group.add_argument(
            '--sort-column',
            action='append',
            default=[],
            dest='sort_columns',
            metavar='SORT_COLUMN',
            help=(
                'specify the column(s) to sort the data (columns specified '
                'first have a priority, non-existing columns are ignored), '
                'can be repeated'
            ),
        )
        sort_dir_group = group.add_mutually_exclusive_group()
        sort_dir_group.add_argument(
            '--sort-ascending',
            action='store_const',
            dest='sort_direction',
            const='asc',
            help=('sort the column(s) in ascending order'),
        )
        sort_dir_group.add_argument(
            '--sort-descending',
            action='store_const',
            dest='sort_direction',
            const='desc',
            help=('sort the column(s) in descending order'),
        )
        return parser

    def produce_output(
        self,
        parsed_args: argparse.Namespace,
        column_names: collections.abc.Sequence[str],
        data: collections.abc.Iterable[collections.abc.Sequence[ty.Any]],
    ) -> int:
        if parsed_args.sort_columns and self.need_sort_by_cliff:
            indexes = [
                column_names.index(c)
                for c in parsed_args.sort_columns
                if c in column_names
            ]
            reverse = parsed_args.sort_direction == 'desc'
            for index in indexes[::-1]:
                try:
                    # We need to handle unset values (i.e. None) so we sort on
                    # multiple conditions: the first comparing the results of
                    # an 'is None' type check and the second comparing the
                    # actual value. The second condition will only be checked
                    # if the first returns True, which only happens if the
                    # returns from the 'is None' check on the two values are
                    # the same, i.e. both None or both not-None
                    data = sorted(
                        data,
                        key=lambda k: (k[index] is None, k[index]),
                        reverse=reverse,
                    )
                except TypeError:
                    # Simply log and then ignore this; sorting is best effort
                    self.log.warning(
                        "Could not sort on field '%s'; unsortable types",
                        parsed_args.sort_columns[index],
                    )

        columns_to_include, selector = self._generate_columns_and_selector(
            parsed_args,
            column_names,
        )
        if selector:
            # Generator expression to only return the parts of a row
            # of data that the user has expressed interest in
            # seeing. We have to convert the compress() output to a
            # list so the table formatter can ask for its length.
            data = (
                list(self._compress_iterable(row, selector)) for row in data
            )

        self.formatter.emit_list(
            columns_to_include,
            data,
            self.app.stdout,
            parsed_args,
        )

        return 0



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from unittest import mock

import testscenarios  # type: ignore

from cliff import command
from cliff import commandmanager
from cliff.tests import base
from cliff.tests import utils


load_tests = testscenarios.load_tests_apply_scenarios


class TestLookupAndFind(base.TestBase):
    scenarios = [
        ('one-word', {'argv': ['one']}),
        ('two-words', {'argv': ['two', 'words']}),
        ('three-words', {'argv': ['three', 'word', 'command']}),
    ]

    def test(self):
        mgr = utils.TestCommandManager(utils.TEST_NAMESPACE)
        cmd, name, remaining = mgr.find_command(self.argv)
        self.assertTrue(cmd)
        self.assertEqual(' '.join(self.argv), name)
        self.assertFalse(remaining)


class TestLookupWithRemainder(base.TestBase):
    scenarios = [
        ('one', {'argv': ['one', '--opt']}),
        ('two', {'argv': ['two', 'words', '--opt']}),
        ('three', {'argv': ['three', 'word', 'command', '--opt']}),
    ]

    def test(self):
        mgr = utils.TestCommandManager(utils.TEST_NAMESPACE)
        cmd, name, remaining = mgr.find_command(self.argv)
        self.assertTrue(cmd)
        self.assertEqual(['--opt'], remaining)


class TestFindInvalidCommand(base.TestBase):
    scenarios = [
        ('no-such-command', {'argv': ['a', '-b']}),
        ('no-command-given', {'argv': ['-b']}),
    ]

    def test(self):
        mgr = utils.TestCommandManager(utils.TEST_NAMESPACE)
        try:
            mgr.find_command(self.argv)
        except ValueError as err:
            # make sure err include 'a' when ['a', '-b']
            self.assertIn(self.argv[0], str(err))
            self.assertIn('-b', str(err))
        else:
            self.fail('expected a failure')


class TestFindUnknownCommand(base.TestBase):
    def test(self):
        mgr = utils.TestCommandManager(utils.TEST_NAMESPACE)
        try:
            mgr.find_command(['a', 'b'])
        except ValueError as err:
            self.assertIn("['a', 'b']", str(err))
        else:
            self.fail('expected a failure')


class FooCommand(command.Command):
    """Description of foo command."""

    def take_action(self, parsed_args):
        return 'foo'


class FooBarCommand(command.Command):
    """Description of foobar command."""

    def take_action(self, parsed_args):
        return 'foobar'


class TestDynamicCommands(base.TestBase):
    def test_add(self):
        mgr = utils.TestCommandManager(utils.TEST_NAMESPACE)
        mock_cmd = mock.Mock()
        mgr.add_command('mock', mock_cmd)
        found_cmd, name, args = mgr.find_command(['mock'])
        self.assertIs(mock_cmd, found_cmd)

    def test_intersected_commands(self):
        mgr = utils.TestCommandManager(utils.TEST_NAMESPACE)
        mgr.add_command('foo', FooCommand)
        mgr.add_command('foo bar', FooBarCommand)

        self.assertIs(FooBarCommand, mgr.find_command(['foo', 'bar'])[0])
        self.assertIs(FooCommand, mgr.find_command(['foo', 'arg0'])[0])


class TestLoad(base.TestBase):
    def test_load_commands(self):
        testcmd = mock.Mock()
        testcmd.name = 'test_cmd'
        testcmd.module_name = 'module.a'
        mock_get_group_all = mock.Mock(return_value=[testcmd])
        with mock.patch(
            'stevedore.ExtensionManager', mock_get_group_all
        ) as mock_manager:
            mgr = commandmanager.CommandManager('test')

        mock_manager.assert_called_once_with('test')
        names = [n for n, v in mgr]
        self.assertEqual(['test cmd'], names)

    def test_load_commands_keep_underscores(self):
        testcmd = mock.Mock()
        testcmd.name = 'test_cmd'
        testcmd.module_name = 'module.a'
        mock_get_group_all = mock.Mock(return_value=[testcmd])
        with mock.patch(
            'stevedore.ExtensionManager', mock_get_group_all
        ) as mock_manager:
            mgr = commandmanager.CommandManager(
                'test',
                convert_underscores=False,
            )

        mock_manager.assert_called_once_with('test')
        names = [n for n, v in mgr]
        self.assertEqual(['test_cmd'], names)

    def test_load_commands_ignore_modules(self):
        testcmd_normal = mock.Mock()
        testcmd_normal.name = 'normal_cmd'
        testcmd_normal.module_name = 'normal.module'
        testcmd_ignored = mock.Mock()
        testcmd_ignored.name = 'ignored_cmd'
        testcmd_ignored.module_name = 'ignored.module'
        mock_get_group_all = mock.Mock(
            return_value=[testcmd_normal, testcmd_ignored]
        )
        with mock.patch(
            'stevedore.ExtensionManager', mock_get_group_all
        ) as mock_manager:
            mgr = commandmanager.CommandManager(
                'test',
                ignored_modules=['ignored.module'],
            )

        mock_manager.assert_called_once_with('test')
        names = [n for n, v in mgr]
        self.assertEqual(['normal cmd'], names)

    def test_load_commands_duplicates(self):
        testcmd_a = mock.Mock()
        testcmd_a.name = 'test_cmd'
        testcmd_a.module_name = 'module.a'
        testcmd_a.entry_point.value = 'module.a:TestCmd'
        testcmd_b = mock.Mock()
        testcmd_b.name = 'test_cmd'
        testcmd_b.module_name = 'module.b'
        testcmd_b.entry_point.value = 'module.b:TestCmd'
        mock_get_group_all = mock.Mock(return_value=[testcmd_a, testcmd_b])
        with mock.patch(
            'stevedore.ExtensionManager', mock_get_group_all
        ) as mock_manager:
            mgr = commandmanager.CommandManager('test')

        mock_manager.assert_called_once_with('test')
        names = [n for n, v in mgr]
        commands = [v for _, v in mgr]
        self.assertEqual(['test cmd'], names)
        # we should have ended up with the second command
        self.assertEqual([testcmd_b.entry_point], commands)
        self.assertIn(
            "found duplicate implementations of the 'test cmd' command",
            self.logger.output,
        )


class FauxCommand(command.Command):
    def take_action(self, parsed_args):
        return 0


class FauxCommand2(FauxCommand):
    pass


class TestLegacyCommand(base.TestBase):
    def test_find_legacy(self):
        mgr = utils.TestCommandManager(utils.TEST_NAMESPACE)
        mgr.add_command('new name', FauxCommand)
        mgr.add_legacy_command('old name', 'new name')
        cmd, name, remaining = mgr.find_command(['old', 'name'])
        self.assertIs(cmd, FauxCommand)
        self.assertEqual(name, 'old name')

    def test_legacy_overrides_new(self):
        mgr = utils.TestCommandManager(utils.TEST_NAMESPACE)
        mgr.add_command('cmd1', FauxCommand)
        mgr.add_command('cmd2', FauxCommand2)
        mgr.add_legacy_command('cmd2', 'cmd1')
        cmd, name, remaining = mgr.find_command(['cmd2'])
        self.assertIs(cmd, FauxCommand)
        self.assertEqual(name, 'cmd2')

    def test_no_legacy(self):
        mgr = utils.TestCommandManager(utils.TEST_NAMESPACE)
        mgr.add_command('cmd1', FauxCommand)
        self.assertRaises(
            ValueError,
            mgr.find_command,
            ['cmd2'],
        )

    def test_no_command(self):
        mgr = utils.TestCommandManager(utils.TEST_NAMESPACE)
        mgr.add_legacy_command('cmd2', 'cmd1')
        self.assertRaises(
            ValueError,
            mgr.find_command,
            ['cmd2'],
        )


class TestLookupAndFindPartialName(base.TestBase):
    scenarios = [
        ('one-word', {'argv': ['o']}),
        ('two-words', {'argv': ['t', 'w']}),
        ('three-words', {'argv': ['t', 'w', 'c']}),
    ]

    def test(self):
        mgr = utils.TestCommandManager(utils.TEST_NAMESPACE)
        cmd, name, remaining = mgr.find_command(self.argv)
        self.assertTrue(cmd)
        self.assertEqual(' '.join(self.argv), name)
        self.assertFalse(remaining)


class TestIsModuleIgnored(base.TestBase):
    def test_match(self):
        result = commandmanager.CommandManager._is_module_ignored(
            'foo.bar.baz', ['foo.bar.baz', 'other.module']
        )
        self.assertTrue(result)
        result = commandmanager.CommandManager._is_module_ignored(
            'foo.bar.baz', ['foo.bar', 'other.module']
        )
        self.assertTrue(result)
        result = commandmanager.CommandManager._is_module_ignored(
            'foo.bar.baz', ['foo', 'other.module']
        )
        self.assertTrue(result)

    def test_no_match(self):
        result = commandmanager.CommandManager._is_module_ignored(
            'foo.bar.baz', ['other.module', 'another.package']
        )
        self.assertFalse(result)

    def test_no_ignores(self):
        result = commandmanager.CommandManager._is_module_ignored(
            'foo.bar.baz', []
        )
        self.assertFalse(result)


class TestGetByPartialName(base.TestBase):
    def setUp(self):
        super().setUp()
        self.commands = {
            'resource provider list': 1,
            'resource class list': 2,
            'server list': 3,
            'service list': 4,
        }

    def test_no_candidates(self):
        self.assertEqual(
            [],
            commandmanager._get_commands_by_partial_name(
                ['r', 'p'], list(self.commands)
            ),
        )
        self.assertEqual(
            [],
            commandmanager._get_commands_by_partial_name(
                ['r', 'p', 'c'], list(self.commands)
            ),
        )

    def test_multiple_candidates(self):
        self.assertEqual(
            2,
            len(
                commandmanager._get_commands_by_partial_name(
                    ['se', 'li'], list(self.commands)
                )
            ),
        )

    def test_one_candidate(self):
        self.assertEqual(
            ['resource provider list'],
            commandmanager._get_commands_by_partial_name(
                ['r', 'p', 'l'], list(self.commands)
            ),
        )
        self.assertEqual(
            ['resource provider list'],
            commandmanager._get_commands_by_partial_name(
                ['resource', 'provider', 'list'], list(self.commands)
            ),
        )
        self.assertEqual(
            ['server list'],
            commandmanager._get_commands_by_partial_name(
                ['serve', 'l'], list(self.commands)
            ),
        )


FAKE_CMD_ONE = commandmanager.EntryPointWrapper('one', utils.TestCommand)
FAKE_CMD_TWO = commandmanager.EntryPointWrapper('two', utils.TestCommand)
FAKE_CMD_ALPHA = commandmanager.EntryPointWrapper('alpha', utils.TestCommand)
FAKE_CMD_BETA = commandmanager.EntryPointWrapper('beta', utils.TestCommand)


class FakeCommandManager(commandmanager.CommandManager):
    commands = {}

    def load_commands(self, namespace):
        if namespace == 'test':
            self.commands['one'] = FAKE_CMD_ONE
            self.commands['two'] = FAKE_CMD_TWO
            self.group_list.append(namespace)
        elif namespace == 'greek':
            self.commands['alpha'] = FAKE_CMD_ALPHA
            self.commands['beta'] = FAKE_CMD_BETA
            self.group_list.append(namespace)


class TestCommandManagerGroups(base.TestBase):
    def test_add_command_group(self):
        mgr = FakeCommandManager('test')

        # Make sure add_command() still functions
        mock_cmd_one = mock.Mock()
        mgr.add_command('mock', mock_cmd_one)
        cmd_mock, name, args = mgr.find_command(['mock'])
        self.assertEqual(mock_cmd_one, cmd_mock)

        # Find a command added in initialization
        cmd_one, name, args = mgr.find_command(['one'])
        self.assertEqual(FAKE_CMD_ONE.load(), cmd_one)

        # Load another command group
        mgr.add_command_group('greek')

        # Find a new command
        cmd_alpha, name, args = mgr.find_command(['alpha'])
        self.assertEqual(FAKE_CMD_ALPHA.load(), cmd_alpha)

        # Ensure that the original commands were not overwritten
        cmd_two, name, args = mgr.find_command(['two'])
        self.assertEqual(FAKE_CMD_TWO.load(), cmd_two)

    def test_get_command_groups(self):
        mgr = FakeCommandManager('test')

        # Make sure add_command() still functions
        mock_cmd_one = mock.Mock()
        mgr.add_command('mock', mock_cmd_one)
        cmd_mock, name, args = mgr.find_command(['mock'])
        self.assertEqual(mock_cmd_one, cmd_mock)

        # Load another command group
        mgr.add_command_group('greek')

        gl = mgr.get_command_groups()
        self.assertEqual(['test', 'greek'], gl)

    def test_get_command_names(self):
        mock_cmd_one = mock.Mock()
        mock_cmd_one.name = 'one'
        mock_cmd_one.module_name = 'module.a'
        mock_cmd_two = mock.Mock()
        mock_cmd_two.name = 'cmd two'
        mock_cmd_two.module_name = 'module.b'
        mock_get_group_all = mock.Mock(
            return_value=[mock_cmd_one, mock_cmd_two],
        )
        with mock.patch(
            'stevedore.ExtensionManager',
            mock_get_group_all,
        ) as mock_manager:
            mgr = commandmanager.CommandManager('test')
            mock_manager.assert_called_once_with('test')

            cmds = mgr.get_command_names('test')
            mock_manager.assert_has_calls(
                [mock.call('test'), mock.call('test')]
            )
            self.assertEqual(['one', 'cmd two'], cmds)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""Bash completion for the CLI."""

import abc
import argparse
import logging
import typing as ty

import stevedore

from cliff import _argparse
from cliff import command as _command

if ty.TYPE_CHECKING:
    from cliff import app


class CompleteDictionary:
    """dictionary for bash completion"""

    def __init__(self) -> None:
        self._dictionary: dict[str, str] = {}

    def add_command(
        self, command: list[str], actions: list[argparse.Action]
    ) -> None:
        optstr = ' '.join(
            opt for action in actions for opt in action.option_strings
        )
        dicto = self._dictionary
        last_cmd = command[-1]
        for subcmd in command[:-1]:
            subdata = dicto.get(subcmd)
            # If there is a string in corresponding dictionary, it means the
            # verb used for the command exists already.
            # For example, {'cmd': 'action'}, and we add the command
            # 'cmd_other'. We want the result to be
            # {'cmd': 'action other', 'cmd_other': 'sub_action'}
            if isinstance(subdata, str):
                subdata += ' ' + last_cmd
                dicto[subcmd] = subdata
                last_cmd = subcmd + '_' + last_cmd
            else:
                # TODO(stephenfin): What is going on here? 😅
                dicto = dicto.setdefault(subcmd, {})  # type: ignore
        dicto[last_cmd] = optstr

    def get_commands(self) -> str:
        return ' '.join(k for k in sorted(self._dictionary.keys()))

    def _get_data_recurse(
        self, dictionary: dict[str, ty.Any], path: str
    ) -> list[tuple[str, str]]:
        ray = []
        keys = sorted(dictionary.keys())
        for cmd in keys:
            name = path + "_" + cmd if path else cmd
            value = dictionary[cmd]
            if isinstance(value, str):
                ray.append((name, value))
            else:
                cmdlist = ' '.join(sorted(value.keys()))
                ray.append((name, cmdlist))
                ray += self._get_data_recurse(value, name)
        return ray

    def get_data(self) -> list[tuple[str, str]]:
        return sorted(self._get_data_recurse(self._dictionary, ""))


class CompleteShellBase(metaclass=abc.ABCMeta):
    """base class for bash completion generation"""

    def __init__(self, name: str, output: ty.TextIO) -> None:
        self.name = str(name)
        self.output = output

    @abc.abstractmethod
    def get_header(self) -> str: ...

    @abc.abstractmethod
    def get_trailer(self) -> str: ...

    def write(self, cmdo: str, data: list[tuple[str, str]]) -> None:
        self.output.write(self.get_header())
        self.output.write(f"  cmds='{cmdo}'\n")
        for datum in data:
            datum = (datum[0].replace('-', '_'), datum[1])
            self.output.write('  cmds_{}=\'{}\'\n'.format(*datum))
        self.output.write(self.get_trailer())

    @property
    def escaped_name(self) -> str:
        return self.name.replace('-', '_')


class CompleteNoCode(CompleteShellBase):
    """completion with no code"""

    def __init__(self, name: str, output: ty.TextIO) -> None:
        super().__init__(name, output)

    def get_header(self) -> str:
        return ''

    def get_trailer(self) -> str:
        return ''


class CompleteBash(CompleteShellBase):
    """completion for bash"""

    def __init__(self, name: str, output: ty.TextIO) -> None:
        super().__init__(name, output)

    def get_header(self) -> str:
        return (
            '_'
            + self.escaped_name
            + """()
{
  local cur prev words
  COMPREPLY=()
  _get_comp_words_by_ref -n : cur prev words

  # Command data:
"""
        )

    def get_trailer(self) -> str:
        return (
            """
  dash=-
  underscore=_
  cmd=""
  words[0]=""
  completed="${cmds}"
  for var in "${words[@]:1}"
  do
    if [[ ${var} == -* ]] ; then
      break
    fi
    if [ -z "${cmd}" ] ; then
      proposed="${var}"
    else
      proposed="${cmd}_${var}"
    fi
    local i="cmds_${proposed}"
    i=${i//$dash/$underscore}
    local comp="${!i}"
    if [ -z "${comp}" ] ; then
      break
    fi
    if [[ ${comp} == -* ]] ; then
      if [[ ${cur} != -* ]] ; then
        completed=""
        break
      fi
    fi
    cmd="${proposed}"
    completed="${comp}"
  done

  if [ -z "${completed}" ] ; then
    COMPREPLY=( $( compgen -f -- "$cur" ) $( compgen -d -- "$cur" ) )
  else
    COMPREPLY=( $(compgen -W "${completed}" -- ${cur}) )
  fi
  return 0
}
complete -F _"""
            + self.escaped_name
            + ' '
            + self.name
            + '\n'
        )


class CompleteCommand(_command.Command):
    """print bash completion command"""

    log = logging.getLogger(__name__ + '.CompleteCommand')
    _formatters: stevedore.ExtensionManager[CompleteShellBase]

    def __init__(
        self,
        app: 'app.App',
        app_args: argparse.Namespace,
        cmd_name: str | None = None,
    ) -> None:
        super().__init__(app, app_args, cmd_name)
        self._formatters = stevedore.ExtensionManager(
            namespace='cliff.formatter.completion',
        )

    def get_parser(self, prog_name: str) -> _argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--name",
            default=None,
            metavar='<command_name>',
            help="Command name to support with command completion",
        )
        parser.add_argument(
            "--shell",
            default='bash',
            metavar='<shell>',
            choices=sorted(self._formatters.names()),
            help="Shell being used. Use none for data only (default: bash)",
        )
        return parser

    def get_actions(self, command: list[str]) -> list[argparse.Action]:
        the_cmd = self.app.command_manager.find_command(command)
        cmd_factory, cmd_name, search_args = the_cmd
        cmd: _command.Command = cmd_factory(self.app, None)
        if self.app.interactive_mode:
            full_name = cmd_name
        else:
            full_name = ' '.join([self.app.NAME, cmd_name])
        cmd_parser = cmd.get_parser(full_name)
        return cmd_parser._get_optional_actions()

    def take_action(self, parsed_args: argparse.Namespace) -> int:
        self.log.debug(f'take_action({parsed_args})')

        name = parsed_args.name or self.app.NAME
        try:
            shell_factory = self._formatters[parsed_args.shell].plugin
        except KeyError:
            raise RuntimeError(f'Unknown shell syntax {parsed_args.shell!r}')
        shell = shell_factory(name, self.app.stdout)

        dicto = CompleteDictionary()
        for cmd in self.app.command_manager:
            command = cmd[0].split()
            dicto.add_command(command, self.get_actions(command))

        shell.write(dicto.get_commands(), dicto.get_data())

        return 0



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""Application base class for displaying data."""

import abc
import argparse
import collections.abc
from itertools import compress
import typing as ty

import stevedore

from cliff import app
from cliff import _argparse
from cliff import command
from cliff.formatters import base as base_formatters

_T = ty.TypeVar("_T")


class DisplayCommandBase(
    command.Command,
    ty.Generic[base_formatters.FormatterT],
    metaclass=abc.ABCMeta,
):
    """Command base class for displaying data about a single object."""

    _formatter_plugins: stevedore.ExtensionManager[base_formatters.FormatterT]
    formatter: base_formatters.FormatterT

    def __init__(
        self,
        app: app.App,
        app_args: argparse.Namespace | None,
        cmd_name: str | None = None,
    ) -> None:
        super().__init__(app, app_args, cmd_name=cmd_name)
        self._formatter_plugins = self._load_formatter_plugins()

    @property
    @abc.abstractmethod
    def formatter_namespace(self) -> str:
        """Namespace to use for loading formatter plugins."""

    @property
    @abc.abstractmethod
    def formatter_default(self) -> str:
        """String specifying the name of the default formatter."""

    def _load_formatter_plugins(
        self,
    ) -> stevedore.ExtensionManager[base_formatters.FormatterT]:
        # Here so tests can override
        return stevedore.ExtensionManager(
            self.formatter_namespace,
            invoke_on_load=True,
        )

    def get_parser(self, prog_name: str) -> _argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        formatter_group = parser.add_argument_group(
            title='output formatters',
            description='output formatter options',
        )
        self._formatter_group = formatter_group
        formatter_choices = sorted(self._formatter_plugins.names())
        formatter_default = self.formatter_default
        if formatter_default not in formatter_choices:
            formatter_default = formatter_choices[0]
        formatter_group.add_argument(
            '-f',
            '--format',
            dest='formatter',
            action='store',
            choices=formatter_choices,
            default=formatter_default,
            help=f'the output format, defaults to {formatter_default}',
        )
        formatter_group.add_argument(
            '-c',
            '--column',
            action='append',
            default=[],
            dest='columns',
            metavar='COLUMN',
            help=(
                'specify the column(s) to include, can be '
                'repeated to show multiple columns'
            ),
        )
        for formatter in self._formatter_plugins:
            assert formatter.obj is not None  # noqa
            formatter.obj.add_argument_group(parser)
        return parser

    @abc.abstractmethod
    def produce_output(
        self,
        parsed_args: argparse.Namespace,
        column_names: collections.abc.Sequence[str],
        data: ty.Any,
    ) -> int:
        """Use the formatter to generate the output.

        :param parsed_args: argparse.Namespace instance with argument values
        :param column_names: sequence of strings containing names
                             of output columns
        :param data: iterable with values matching the column names
        :returns: a status code
        """

    def _generate_columns_and_selector(
        self,
        parsed_args: argparse.Namespace,
        column_names: collections.abc.Sequence[str],
    ) -> tuple[list[str], list[bool] | None]:
        """Generate included columns and selector according to parsed args.

        We normalize the column names so that someone can do e.g. '-c
        server_name' when the output field is actually called 'Server Name'.

        :param parsed_args: argparse.Namespace instance with argument values
        :param column_names: sequence of strings containing names
                             of output columns
        """
        if not parsed_args.columns:
            return list(column_names), None

        def normalize_column(column_name: str) -> str:
            return column_name.lower().strip().replace(' ', '_')

        requested_columns = [normalize_column(c) for c in parsed_args.columns]
        columns_to_include = [
            c for c in column_names if normalize_column(c) in requested_columns
        ]
        if not columns_to_include:
            raise ValueError(
                f'No recognized column names in {str(parsed_args.columns)}. '
                f'Recognized columns are {str(column_names)}.'
            )

        # Set up argument to compress()
        selector = [(c in columns_to_include) for c in column_names]
        return columns_to_include, selector

    def run(self, parsed_args: argparse.Namespace) -> int:
        parsed_args = self._run_before_hooks(parsed_args)
        formatter = self._formatter_plugins[parsed_args.formatter].obj
        assert formatter is not None  # noqa
        self.formatter = formatter
        column_names, data = self.take_action(parsed_args)
        column_names, data = self._run_after_hooks(
            parsed_args, (column_names, data)
        )
        self.produce_output(parsed_args, column_names, data)
        return 0

    def _run_after_hooks(  # type: ignore[override]
        self,
        parsed_args: argparse.Namespace,
        data: tuple[
            collections.abc.Sequence[str], collections.abc.Iterable[ty.Any]
        ],
    ) -> tuple[
        collections.abc.Sequence[str], collections.abc.Iterable[ty.Any]
    ]:
        """Calls after() method of the hooks.

        This method is intended to be called from the run() method after
        take_action() is called.

        This method should only be overridden by developers creating new
        command base classes and only if it is necessary to have different
        hook processing behavior.
        """
        for hook in self._hooks:
            # we need to ignore the types since CommandHook states that it
            # should return an integer, but they'll actually return a tuple of
            # column names and data when used with DisplayCommandBase
            # subclasses
            ret = hook.obj.after(parsed_args, data)  # type: ignore
            # If the return is None do not change return_code, otherwise
            # set up to pass it to the next hook
            if ret is not None:
                data = ret  # type: ignore
        return data

    @staticmethod
    def _compress_iterable(
        iterable: collections.abc.Iterable[_T],
        selectors: collections.abc.Iterable[ty.Any],
    ) -> collections.abc.Iterator[_T]:
        return compress(iterable, selectors)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""Output formatters for JSON."""

import argparse
import collections.abc
import json
import typing as ty

from cliff import columns
from cliff.formatters import base


class JSONFormatter(base.ListFormatter, base.SingleFormatter):
    def add_argument_group(self, parser: argparse.ArgumentParser) -> None:
        group = parser.add_argument_group(title='json formatter')
        group.add_argument(
            '--noindent',
            action='store_true',
            dest='noindent',
            help='whether to disable indenting the JSON',
        )

    def emit_list(
        self,
        column_names: collections.abc.Sequence[str],
        data: collections.abc.Iterable[collections.abc.Sequence[ty.Any]],
        stdout: ty.TextIO,
        parsed_args: argparse.Namespace,
    ) -> None:
        items = []
        for item in data:
            items.append(
                {
                    n: (
                        i.machine_readable()
                        if isinstance(i, columns.FormattableColumn)
                        else i
                    )
                    for n, i in zip(column_names, item)
                }
            )
        indent = None if parsed_args.noindent else 2
        json.dump(items, stdout, indent=indent)
        stdout.write('\n')

    def emit_one(
        self,
        column_names: collections.abc.Sequence[str],
        data: collections.abc.Sequence[ty.Any],
        stdout: ty.TextIO,
        parsed_args: argparse.Namespace,
    ) -> None:
        one = {
            n: (
                i.machine_readable()
                if isinstance(i, columns.FormattableColumn)
                else i
            )
            for n, i in zip(column_names, data)
        }
        indent = None if parsed_args.noindent else 2
        json.dump(one, stdout, indent=indent)
        stdout.write('\n')



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""Application base class for displaying data about a single object."""

import abc
import argparse
import collections.abc
import typing as ty

from cliff import display
from cliff.formatters import base as base_formatters


class ShowOne(
    display.DisplayCommandBase[base_formatters.SingleFormatter],
    metaclass=abc.ABCMeta,
):
    """Command base class for displaying data about a single object."""

    @property
    def formatter_namespace(self) -> str:
        return 'cliff.formatter.show'

    @property
    def formatter_default(self) -> str:
        return 'table'

    @abc.abstractmethod
    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[
        collections.abc.Sequence[str], collections.abc.Iterable[ty.Any]
    ]:
        """Run command.

        Return a tuple containing the column names and an iterable containing
        the data to be listed.
        """

    def produce_output(
        self,
        parsed_args: argparse.Namespace,
        column_names: collections.abc.Sequence[str],
        data: collections.abc.Sequence[ty.Any],
    ) -> int:
        columns_to_include, selector = self._generate_columns_and_selector(
            parsed_args, column_names
        )
        if selector:
            data = list(self._compress_iterable(data, selector))
        self.formatter.emit_one(
            columns_to_include, data, self.app.stdout, parsed_args
        )
        return 0

    def dict2columns(
        self, data: dict[str, ty.Any]
    ) -> tuple[tuple[str, ...], tuple[ty.Any, ...]]:
        """Implement the common task of converting a dict-based object
        to the two-column output that ShowOne expects.
        """
        if not data:
            return ((), ())
        else:
            return (tuple(data.keys()), tuple(data.values()))



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""Overrides of standard argparse behavior."""

import argparse
import collections.abc
import typing as ty
import warnings

import autopage.argparse


class ArgumentParser(autopage.argparse.ArgumentParser):
    # NOTE(dhellmann): We have to override the methods for creating
    # groups to return our objects that know how to deal with the
    # special conflict handler.

    def add_argument_group(
        self, *args: ty.Any, **kwargs: ty.Any
    ) -> '_ArgumentGroup':
        group = _ArgumentGroup(self, *args, **kwargs)
        self._action_groups.append(group)
        return group

    def add_mutually_exclusive_group(
        self, **kwargs: ty.Any
    ) -> '_MutuallyExclusiveGroup':
        group = _MutuallyExclusiveGroup(self, **kwargs)
        self._mutually_exclusive_groups.append(group)
        return group

    def _handle_conflict_ignore(
        self,
        action: argparse.Action,
        conflicting_actions: collections.abc.Iterable[
            tuple[str, argparse.Action]
        ],
    ) -> None:
        _handle_conflict_ignore(
            self,
            self._option_string_actions,
            action,
            conflicting_actions,
        )


def _handle_conflict_ignore(
    container: argparse._ActionsContainer,
    option_string_actions: dict[str, argparse.Action],
    new_action: argparse.Action,
    conflicting_actions: collections.abc.Iterable[tuple[str, argparse.Action]],
) -> None:
    # Remember the option strings the new action starts with so we can
    # restore them as part of error reporting if we need to.
    original_option_strings = new_action.option_strings

    # Remove all of the conflicting option strings from the new action
    # and report an error if none are left at the end.
    for option_string, action in conflicting_actions:
        # remove the conflicting option from the new action
        new_action.option_strings.remove(option_string)  # type: ignore
        warnings.warn(
            f'Ignoring option string {option_string} for new action '
            f'because it conflicts with an existing option.'
        )

        # if the option now has no option string, remove it from the
        # container holding it
        if not new_action.option_strings:
            new_action.option_strings = original_option_strings
            raise argparse.ArgumentError(
                new_action,
                'Cannot resolve conflicting option string, '
                'all names conflict.',
            )


class _ArgumentGroup(argparse._ArgumentGroup):
    # NOTE(dhellmann): We have to override the methods for creating
    # groups to return our objects that know how to deal with the
    # special conflict handler.

    def add_argument_group(
        self, *args: ty.Any, **kwargs: ty.Any
    ) -> '_ArgumentGroup':
        group = _ArgumentGroup(self, *args, **kwargs)
        self._action_groups.append(group)
        return group

    def add_mutually_exclusive_group(
        self, **kwargs: ty.Any
    ) -> '_MutuallyExclusiveGroup':
        group = _MutuallyExclusiveGroup(self, **kwargs)
        self._mutually_exclusive_groups.append(group)
        return group

    def _handle_conflict_ignore(
        self,
        action: argparse.Action,
        conflicting_actions: collections.abc.Iterable[
            tuple[str, argparse.Action]
        ],
    ) -> None:
        _handle_conflict_ignore(
            self,
            self._option_string_actions,
            action,
            conflicting_actions,
        )


class _MutuallyExclusiveGroup(argparse._MutuallyExclusiveGroup):
    # NOTE(dhellmann): We have to override the methods for creating
    # groups to return our objects that know how to deal with the
    # special conflict handler.

    def add_argument_group(
        self, *args: ty.Any, **kwargs: ty.Any
    ) -> '_ArgumentGroup':
        group = _ArgumentGroup(self, *args, **kwargs)
        self._action_groups.append(group)
        return group

    def add_mutually_exclusive_group(
        self, **kwargs: ty.Any
    ) -> '_MutuallyExclusiveGroup':
        group = _MutuallyExclusiveGroup(self, **kwargs)
        self._mutually_exclusive_groups.append(group)
        return group

    def _handle_conflict_ignore(
        self,
        action: argparse.Action,
        conflicting_actions: collections.abc.Iterable[
            tuple[str, argparse.Action]
        ],
    ) -> None:
        _handle_conflict_ignore(
            self,
            self._option_string_actions,
            action,
            conflicting_actions,
        )


class SmartHelpFormatter(autopage.argparse.HelpFormatter):
    """Smart help formatter to output raw help message if help contain \\n.

    Some command help messages maybe have multiple line content, the built-in
    argparse.HelpFormatter wrap and split the content according to width, and
    ignore \\n in the raw help message, it merge multiple line content in one
    line to output, that looks messy. SmartHelpFormatter keep the raw help
    message format if it contain \\n, and wrap long line like HelpFormatter
    behavior.
    """

    def _split_lines(self, text: str, width: int) -> list[str]:
        lines = text.splitlines() if '\n' in text else [text]
        wrap_lines = []
        for each_line in lines:
            wrap_lines.extend(super()._split_lines(each_line, width))
        return wrap_lines



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import argparse
import typing as ty
import weakref

from unittest import mock

from cliff.formatters import base as base_formatters
from cliff import lister
from cliff.tests import base


class FauxFormatter(base_formatters.ListFormatter):
    def __init__(self):
        self.args = []
        self.obj = weakref.proxy(self)

    def add_argument_group(self, parser: argparse.ArgumentParser) -> None:
        return None

    def emit_list(self, columns, data, stdout, args):
        self.args.append((columns, data))


class ExerciseLister(lister.Lister):
    data: list[tuple[ty.Any, ty.Any]] = [('a', 'A'), ('b', 'B'), ('c', 'A')]

    def _load_formatter_plugins(self):
        return {
            'test': FauxFormatter(),
        }

    def take_action(self, parsed_args):
        return (parsed_args.columns, self.data)


class ExerciseListerCustomSort(ExerciseLister):
    need_sort_by_cliff = False


class ExerciseListerNullValues(ExerciseLister):
    data = ExerciseLister.data + [(None, None)]


class ExerciseListerDifferentTypes(ExerciseLister):
    data = ExerciseLister.data + [(1, 0)]


class TestLister(base.TestBase):
    def test_formatter_args(self):
        app = mock.Mock()
        test_lister = ExerciseLister(app, None)

        parsed_args = mock.Mock()
        parsed_args.columns = ('Col1', 'Col2')
        parsed_args.formatter = 'test'
        parsed_args.sort_columns = []

        test_lister.run(parsed_args)

        f = test_lister._formatter_plugins['test']
        assert isinstance(f, FauxFormatter)
        self.assertEqual(1, len(f.args))
        args = f.args[0]
        self.assertEqual(list(parsed_args.columns), args[0])
        data = list(args[1])
        self.assertEqual([['a', 'A'], ['b', 'B'], ['c', 'A']], data)

    def test_filter_by_columns_invalid(self):
        test_lister = ExerciseLister(mock.Mock(), None)
        parsed_args = mock.Mock()
        parsed_args.columns = ('no_exist_column',)
        parsed_args.formatter = 'test'
        parsed_args.sort_columns = []
        with mock.patch.object(test_lister, 'take_action') as mock_take_action:
            mock_take_action.return_value = (('Col1', 'Col2', 'Col3'), [])
            self.assertRaises(
                ValueError,
                test_lister.run,
                parsed_args,
            )

    def test_filter_by_columns_normalized(self):
        test_lister = ExerciseLister(mock.Mock(), None)
        parsed_args = mock.Mock()
        parsed_args.columns = ('col1', 'COL2')
        parsed_args.formatter = 'test'
        parsed_args.sort_columns = []

        test_lister.run(parsed_args)

        f = test_lister._formatter_plugins['test']
        assert isinstance(f, FauxFormatter)
        args = f.args[0]
        data = list(args[1])
        self.assertEqual([['a', 'A'], ['b', 'B'], ['c', 'A']], data)

    def test_sort_by_column_cliff_side_procedure(self):
        test_lister = ExerciseLister(mock.Mock(), None)
        parsed_args = mock.Mock()
        parsed_args.columns = ('Col1', 'Col2')
        parsed_args.formatter = 'test'
        parsed_args.sort_columns = ['Col2', 'Col1']

        test_lister.run(parsed_args)

        f = test_lister._formatter_plugins['test']
        assert isinstance(f, FauxFormatter)
        args = f.args[0]
        data = list(args[1])
        self.assertEqual([['a', 'A'], ['c', 'A'], ['b', 'B']], data)

    def test_sort_by_column_reverse_order(self):
        test_lister = ExerciseLister(mock.Mock(), None)
        parsed_args = mock.Mock()
        parsed_args.columns = ('Col1', 'Col2')
        parsed_args.formatter = 'test'
        parsed_args.sort_columns = ['Col2', 'Col1']
        parsed_args.sort_direction = 'desc'

        test_lister.run(parsed_args)

        f = test_lister._formatter_plugins['test']
        assert isinstance(f, FauxFormatter)
        args = f.args[0]
        data = list(args[1])
        self.assertEqual([['b', 'B'], ['c', 'A'], ['a', 'A']], data)

    def test_sort_by_column_data_already_sorted(self):
        test_lister = ExerciseListerCustomSort(mock.Mock(), None)
        parsed_args = mock.Mock()
        parsed_args.columns = ('Col1', 'Col2')
        parsed_args.formatter = 'test'
        parsed_args.sort_columns = ['Col2', 'Col1']

        test_lister.run(parsed_args)

        f = test_lister._formatter_plugins['test']
        assert isinstance(f, FauxFormatter)
        args = f.args[0]
        data = list(args[1])
        self.assertEqual([['a', 'A'], ['b', 'B'], ['c', 'A']], data)

    def test_sort_by_column_with_null(self):
        test_lister = ExerciseListerNullValues(mock.Mock(), None)
        parsed_args = mock.Mock()
        parsed_args.columns = ('Col1', 'Col2')
        parsed_args.formatter = 'test'
        parsed_args.sort_columns = ['Col2', 'Col1']

        test_lister.run(parsed_args)

        f = test_lister._formatter_plugins['test']
        assert isinstance(f, FauxFormatter)
        args = f.args[0]
        data = list(args[1])
        self.assertEqual(
            [['a', 'A'], ['c', 'A'], ['b', 'B'], [None, None]], data
        )

    def test_sort_by_column_with_different_types(self):
        test_lister = ExerciseListerDifferentTypes(mock.Mock(), None)
        parsed_args = mock.Mock()
        parsed_args.columns = ('Col1', 'Col2')
        parsed_args.formatter = 'test'
        parsed_args.sort_columns = ['Col2', 'Col1']

        with mock.patch.object(lister.Lister, 'log') as mock_log:
            test_lister.run(parsed_args)

        f = test_lister._formatter_plugins['test']
        assert isinstance(f, FauxFormatter)
        args = f.args[0]
        data = list(args[1])
        # The output should be unchanged
        self.assertEqual([['a', 'A'], ['b', 'B'], ['c', 'A'], [1, 0]], data)
        # but we should have logged a warning
        mock_log.warning.assert_has_calls(
            [
                mock.call(
                    "Could not sort on field '%s'; unsortable types", col
                )
                for col in parsed_args.sort_columns
            ]
        )

    def test_sort_by_non_displayed_column(self):
        test_lister = ExerciseLister(mock.Mock(), None)
        parsed_args = mock.Mock()
        parsed_args.columns = ('Col1',)
        parsed_args.formatter = 'test'
        parsed_args.sort_columns = ['Col2']

        with mock.patch.object(test_lister, 'take_action') as mock_take_action:
            mock_take_action.return_value = (
                ('Col1', 'Col2'),
                [['a', 'A'], ['b', 'B'], ['c', 'A']],
            )
            test_lister.run(parsed_args)

        f = test_lister._formatter_plugins['test']
        assert isinstance(f, FauxFormatter)
        args = f.args[0]
        data = list(args[1])
        self.assertEqual([['a'], ['c'], ['b']], data)

    def test_sort_by_non_existing_column(self):
        test_lister = ExerciseLister(mock.Mock(), None)
        parsed_args = mock.Mock()
        parsed_args.columns = ('Col1', 'Col2')
        parsed_args.formatter = 'test'
        parsed_args.sort_columns = ['no_exist_column']

        test_lister.run(parsed_args)

        f = test_lister._formatter_plugins['test']
        assert isinstance(f, FauxFormatter)
        args = f.args[0]
        data = list(args[1])
        self.assertEqual([['a', 'A'], ['b', 'B'], ['c', 'A']], data)



import logging
import os

from cliff.show import ShowOne


class File(ShowOne):
    "Show details about a file"

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument('filename', nargs='?', default='.')
        return parser

    def take_action(self, parsed_args):
        stat_data = os.stat(parsed_args.filename)
        columns = (
            'Name',
            'Size',
            'UID',
            'GID',
            'Modified Time',
        )
        data = (
            parsed_args.filename,
            stat_data.st_size,
            stat_data.st_uid,
            stat_data.st_gid,
            stat_data.st_mtime,
        )
        return (columns, data)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import abc
import argparse

from cliff import _argparse
from cliff import command


class CommandHook(metaclass=abc.ABCMeta):
    """Base class for command hooks.

    Hook methods are executed in the following order:

    1. :meth:`get_epilog`
    2. :meth:`get_parser`
    3. :meth:`before`
    4. :meth:`after`

    :param command: Command instance being invoked
    """

    def __init__(self, command: command.Command):
        self.cmd = command

    @abc.abstractmethod
    def get_parser(
        self, parser: _argparse.ArgumentParser
    ) -> _argparse.ArgumentParser | None:
        """Modify the command :class:`argparse.ArgumentParser`.

        The provided parser is modified in-place, and the return value is not
        used.

        :param parser: An existing ArgumentParser instance to be modified.
        :returns: ArgumentParser or None
        """
        return parser

    @abc.abstractmethod
    def get_epilog(self) -> str | None:
        """Return text to add to the command help epilog.

        :returns: An epilog string or None.
        """
        return ''

    @abc.abstractmethod
    def before(self, parsed_args: argparse.Namespace) -> argparse.Namespace:
        """Called before the command's take_action() method.

        :param parsed_args: The arguments to the command.
        :returns: argparse.Namespace
        """
        return parsed_args

    @abc.abstractmethod
    def after(self, parsed_args: argparse.Namespace, return_code: int) -> int:
        """Called after the command's take_action() method.

        :param parsed_args: The arguments to the command.
        :param return_code: The value returned from take_action().
        :paramtype return_code: int
        :returns: int
        """
        return return_code



#!/usr/bin/env python

from setuptools import find_packages
from setuptools import setup

PROJECT = 'cliffdemo'

# Change docs/sphinx/conf.py too!
VERSION = '0.1'

try:
    long_description = open('README.rst').read()
except OSError:
    long_description = ''

setup(
    name=PROJECT,
    version=VERSION,
    description='Demo app for cliff',
    long_description=long_description,
    author='Doug Hellmann',
    author_email='doug.hellmann@gmail.com',
    url='https://github.com/openstack/cliff',
    download_url='https://github.com/openstack/cliff/tarball/master',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Intended Audience :: Developers',
        'Environment :: Console',
    ],
    platforms=['Any'],
    scripts=[],
    provides=[],
    install_requires=['cliff'],
    namespace_packages=[],
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': ['cliffdemo = cliffdemo.main:main'],
        'cliff.demo': [
            'simple = cliffdemo.simple:Simple',
            'two_part = cliffdemo.simple:Simple',
            'error = cliffdemo.simple:Error',
            'list files = cliffdemo.list:Files',
            'files = cliffdemo.list:Files',
            'file = cliffdemo.show:File',
            'show file = cliffdemo.show:File',
            'unicode = cliffdemo.encoding:Encoding',
            'hooked = cliffdemo.hook:Hooked',
        ],
        'cliff.demo.hooked': [
            'sample-hook = cliffdemo.hook:Hook',
        ],
    },
    zip_safe=False,
)
